@@ -1,7 +1,7 @@
 package Web::Async::WebSocket::Server;
 use Myriad::Class extends => 'IO::Async::Notifier';
 
-our $VERSION = '0.001'; ## VERSION
+our $VERSION = '0.002'; ## VERSION
 ## AUTHORITY
 
 =head1 NAME
@@ -25,9 +25,14 @@ field $ryu : reader : param = undef;
 field $port : reader : param = undef;
 
 field $incoming_client : reader : param = undef;
+field $disconnecting_client : reader : param = undef;
+field $closing_client : reader : param = undef;
+field $active_client : reader { +{ } }
 
 field $on_handshake_failure : reader : param = undef;
 
+field $listening : reader = undef;
+
 method configure (%args) {
     $port = delete $args{port} if exists $args{port};
     $on_handshake_failure = delete $args{on_handshake_failure} if exists $args{on_handshake_failure};
@@ -39,16 +44,17 @@ method _add_to_loop ($loop) {
         $ryu = Ryu::Async->new
     ) unless $ryu;
     $incoming_client //= $self->ryu->source;
+    $closing_client //= $self->ryu->source;
     $self->add_child(
         $srv = IO::Async::Listener->new(
             on_stream => $self->curry::weak::on_stream,
         )
     );
     $self->adopt_future(
-        $srv->listen(
+        $listening = $srv->listen(
             service  => $port,
             socktype => 'stream',
-        )
+        )->on_ready(sub { undef $listening })
     );
 }
 
@@ -64,12 +70,28 @@ method on_stream ($listener, $stream, @) {
             on_handshake_failure => $on_handshake_failure,
         )
     );
+    $active_client->{$client} = $client;
     $incoming_client->emit($client);
     $self->adopt_future(
         $client->handle_connection
     );
 }
 
+method on_client_close ($client, %args) {
+    $closing_client->emit({
+        client => $client,
+        %args,
+    });
+    delete $active_client->{$client} or $log->errorf('Client %s was not recorded', "$client");
+    return;
+}
+
+method on_client_disconnect ($client, @) {
+    $disconnecting_client->emit($client);
+    delete $active_client->{$client} or $log->errorf('Client %s was not recorded', "$client");
+    return;
+}
+
 1;
 
 =head1 AUTHOR
