@@ -1,7 +1,7 @@
 package Web::Async::WebSocket::Server::Connection;
 use Myriad::Class extends => 'IO::Async::Notifier';
 
-our $VERSION = '0.001'; ## VERSION
+our $VERSION = '0.002'; ## VERSION
 ## AUTHORITY
 
 use Web::Async::WebSocket::Frame;
@@ -12,20 +12,28 @@ use POSIX ();
 use Time::Moment;
 use Digest::SHA qw(sha1);
 use MIME::Base64 qw(encode_base64);
+use Unicode::UTF8 qw(valid_utf8);
 
 # As defined in the RFC - it's used as part of the hashing for the security header in the response
 use constant WEBSOCKET_GUID => '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
 
 # Opcodes have a registry here: https://www.iana.org/assignments/websocket/websocket.xhtml#opcode
-my %OPCODE_BY_CODE = (
-    0 => 'continuation',
-    1 => 'text',
-    2 => 'binary',
-    8 => 'close',
-    9 => 'ping',
+our %OPCODE_BY_CODE = (
+    0  => 'continuation',
+    1  => 'text',
+    2  => 'binary',
+    8  => 'close',
+    9  => 'ping',
     10 => 'pong',
 );
-my %OPCODE_BY_NAME = reverse %OPCODE_BY_CODE;
+our %OPCODE_BY_NAME = reverse %OPCODE_BY_CODE;
+
+our %COMPRESSIBLE_OPCODE = (
+    $OPCODE_BY_NAME{text}   => 1,
+    $OPCODE_BY_NAME{binary} => 1,
+);
+
+field $server : reader : param = undef;
 
 # Given the state of websockets in general, this is unlikely to change from `HTTP/1.1` anytime soon
 field $http_version : reader : param = 'HTTP/1.1';
@@ -37,7 +45,11 @@ field $msg : reader : param = 'Switching Protocols';
 # lists just two of 'em
 field $supported_extension : reader : param {
     +{
-        'permessage-deflate' => 1
+        'permessage-deflate' => 1,
+        'server_no_context_takeover' => 1,
+        'client_no_context_takeover' => 1,
+        'server_max_window_bits' => 1,
+        'client_max_window_bits' => 1,
     }
 }
 
@@ -62,13 +74,15 @@ field $on_handshake_failure : param : reader = undef;
 field $incoming_frame : reader : param { $self->ryu->source }
 # A Ryu::Source representing the messages to be sent to the client
 field $outgoing_frame : reader : param { $self->ryu->source }
-# A Future which will resolve with an error if the handshake failed
-field $handshake_failure : reader = undef;
+
+field $compression_options : reader { +{ } }
 
 # The IO::Async::Stream representing the network connection
 # to the client
 field $stream;
 
+field $closed : reader : param = undef;
+
 method configure (%args) {
     $http_version = delete $args{http_version} if exists $args{http_version};
     $status = delete $args{status} if exists $args{status};
@@ -85,11 +99,102 @@ method _add_to_loop ($loop) {
     $on_handshake_failure //= async method ($stream, $error, @) {
         await $stream->write("$http_version 400 $error\x0D\x0A\x0D\x0A");
     };
+    $closed //= $self->loop->new_future;
+    $stream->configure(
+        on_closed => $self->$curry::weak(async method (@) {
+            $closed->done unless $closed->is_ready;
+            $server->on_client_disconnect($self);
+        }),
+    );
+}
+
+=head2 send_text_frame
+
+Send a text frame.
+
+Expects a Unicode Perl text string as the first parameter - this will be
+encoded to UTF-8 and sent to the client.
+
+=cut
+
+async method send_text_frame ($text, %args) {
+    return await $self->write_frame(
+        payload => $text,
+        type => 'text',
+        %args
+    );
+}
+
+=head2 send_binary_frame
+
+Send a binary data frame.
+
+Expects the raw binary data bytes as the first parameter.
+
+=cut
+
+async method send_data_frame ($data, %args) {
+    return await $self->write_frame(
+        payload => $data,
+        type    => 'binary',
+        %args
+    );
+}
+
+=head2 write_frame
+
+Sends one or more frames to the client.
+
+=cut
+
+async method write_frame (%args) {
+    die 'already closed' if $closed->is_ready;
+    for my $frame ($self->prepare_frames(%args)) {
+        await $stream->write($frame);
+    }
+    return;
+}
+
+async method prepare_frames (%args) {
+    my @frames;
+    $log->tracef('Write frame with %s', \%args);
+    my $opcode = $OPCODE_BY_NAME{$args{type}} // die 'invalid frame type';
+    my $compressed = ($args{compress} // 1) && $compression_options->{compress} && $COMPRESSIBLE_OPCODE{$opcode};
+    my $payload = $args{payload};
+    $payload = encode_utf8($payload) if $opcode == $OPCODE_BY_NAME{text};
+
+    $opcode |= 0x80;
+    if($compressed) {
+        $opcode |= 0x40;
+        my $original = length $payload;
+        $payload = $self->deflate($payload);
+        # Strip terminator if we have one
+        $payload =~ s{\x00\x00\xFF\xFF$}{};
+        $log->tracef(
+            'Size after deflation is %d/%d, ratio of %4.1f%%',
+            length($payload),
+            $original,
+            100.0 * (length($payload) / ($original || 1)),
+        );
+    }
+    my $len = length $payload;
+    my $msg = pack('C1', $opcode);
+    if($len < 126) {
+        $msg .= pack('C1', $len);
+    } elsif($len <= 0xFFFF) {
+        $msg .= pack('C1n1', 126, $len);
+    } else {
+        $msg .= pack('C1Q>1', 127, $len);
+    }
+    $msg .= $payload;
+    push @frames, $msg;
+    return @frames;
 }
 
 method deflate ($data) {
+    undef $deflation unless $compression_options->{server_context};
     $deflation //= deflateInit(
-        -WindowBits => -MAX_WBITS
+        -WindowBits => -($compression_options->{server_bits} || 15)
     ) or die "Cannot create a deflation stream\n" ;
 
     my ($output, $status) = $deflation->deflate($data);
@@ -101,8 +206,9 @@ method deflate ($data) {
 }
 
 method inflate ($data) {
+    undef $inflation unless $compression_options->{cilent_context};
     $inflation //= inflateInit(
-        -WindowBits => -MAX_WBITS
+        -WindowBits => -($compression_options->{client_bits} || 15)
     ) or die "Cannot create a deflation stream\n" ;
 
     my ($block, $status) = $inflation->inflate($data);
@@ -151,8 +257,46 @@ async method handle_connection () {
         );
         $output{'Sec-WebSocket-Accept'} = $self->generate_response_key($hdr->{sec_websocket_key});
 
-        my @extensions = grep { $supported_extension->{$_} } map { /([^=]+)/ } split /\s*;\s*/, $hdr->{sec_websocket_extensions};
-        $output{'Sec-Websocket-Extensions'} = join ';', sort @extensions;
+        if(exists $hdr->{sec_websocket_extensions}) {
+            my $extensions;
+            VALID: {
+                SELECTION:
+                for my $selection (split /\s*,\s*/, $hdr->{sec_websocket_extensions} // '') {
+                    my @options = map {; /^(\S+)(?:\s*=\s*(.*)\s*)?$/ ? ($1, $2) : () } split /\s*;\s*/, $selection;
+                    my @order = pairmap { $a } @options;
+                    my %options = @options;
+                    my @invalid = grep { !$supported_extension->{$_} } sort keys %options;
+                    if(@invalid) {
+                        $log->infof('Rejecting invalid option combination %s', \@invalid);
+                        next SELECTION;
+                    }
+
+                    $log->infof('Acceptable options: %s', \%options);
+                    $options{client_max_window_bits} //= 15 if exists $options{client_max_window_bits};
+                    $compression_options->{client_bits} = $options{client_max_window_bits};
+                    $compression_options->{server_bits} = $options{server_max_window_bits} || 15;
+                    $extensions = join '; ', map { defined($options{$_}) ? "$_=$options{$_}" : $_ } @order;
+                    $compression_options->{server_context} = (exists $options{server_no_context_takeover}) ? 0 : 1;
+                    $compression_options->{client_context} = (exists $options{client_no_context_takeover}) ? 0 : 1;
+                    $compression_options->{compress} = 1 if exists $options{'permessage-deflate'};
+                    last VALID;
+                }
+                $log->infof('No acceptable extension options, giving up: %s', $hdr->{sec_websocket_extensions});
+                await $stream->write(
+                    join(
+                        "\x0D\x0A",
+                        "$http_version 400 No acceptable extensions",
+                        (pairmap {
+                            encode_utf8("$a: $b")
+                        } %output),
+                        # Blank line at the end of the headers
+                        '', ''
+                    )
+                );
+                die 'no acceptable extensions';
+            }
+            $output{'Sec-Websocket-Extensions'} = $extensions;
+        }
 
         # Send the entire header block in a single write
         await $stream->write(
@@ -172,6 +316,10 @@ async method handle_connection () {
         return;
     }
 
+    # Once the handshake is complete, we don't need the handler any more,
+    # and keeping it around could lead to unwanted refcount cycles
+    undef $on_handshake_failure;
+
     # Body processing
     try {
         $log->tracef('Start reading frames');
@@ -180,14 +328,13 @@ async method handle_connection () {
             my $frame = await $self->read_frame();
             $log->tracef('Had frame: %s', $frame);
             $incoming_frame->emit($frame);
-#            await $self->write_frame(
-#                type    => 'text',
-#                payload => $payload
-#            );
         }
     } catch ($e) {
-        $log->errorf('Problem, %s', $e);
-        $stream->close;
+        $log->errorf('Problem, %s', $e) unless $e =~ /^EOF/;
+        await $self->close(
+            code   => 1011, # internal error
+            reason => 'Internal error'
+        );
     }
 }
 
@@ -207,8 +354,20 @@ async method read_frame () {
         $len &= ~0x80;
         $fin = ($opcode & 0x80) ? 1 : 0;
         my @rsv = map { ($opcode & $_) ? 1 : 0 } 0x40, 0x20, 0x10;
-        $compressed //= $rsv[0];
+        $compressed //= $compression_options->{compress} && $rsv[0];
+        return await $self->close(
+            code => 1002,
+            reason => 'Reserved bit 0 set with compression disabled',
+        ) if $rsv[0] and not $compression_options->{compress};
+        return await $self->close(
+            code => 1002,
+            reason => 'Unexpected reserved bit set',
+        ) if any { $_ } @rsv;
         $type //= $opcode & 0x0F;
+        return await $self->close(
+            code   => 1002,
+            reason => 'Unknown opcode',
+        ) unless $OPCODE_BY_CODE{$type};
         if($len == 126) {
             ($chunk, $eof) = await $stream->read_exactly(2);
             die "EOF\n" if $eof;
@@ -246,48 +405,55 @@ async method read_frame () {
     } until $fin;
     $data = $self->inflate($data . "\x00\x00\xFF\xFF") if $compressed;
     $log->tracef('Frame opcode is %s', $OPCODE_BY_CODE{$type});
-    $data = decode_utf8($data) if $type == $OPCODE_BY_NAME{text};
+    if($type == $OPCODE_BY_NAME{text}) {
+        return await $self->close(
+            code   => 1002,
+            reason => 'Invalid UTF-8 data in text frame',
+        ) unless valid_utf8($data);
+        $data = decode_utf8($data);
+    }
     $log->tracef('Finished, data is now %s', $data);
-    return Web::Async::WebSocket::Frame->new(
+    my $frame = Web::Async::WebSocket::Frame->new(
         payload => $data,
         opcode => $type
     );
-}
-
-async method write_frame (%args) {
-    my $compressed = $args{compress} // 1;
-    $log->tracef('Write frame with %s', \%args);
-    # FIN
-    my $opcode = $OPCODE_BY_NAME{$args{type}};
-    my $payload = $args{payload};
-    $payload = encode_utf8($payload) if $opcode == $OPCODE_BY_NAME{text};
-
-    $opcode |= 0x80;
-    if($compressed) {
-        $opcode |= 0x40;
-        my $original = length $payload;
-        $payload = $self->deflate($payload);
-        # Strip terminator if we have one
-        $payload =~ s{\x00\x00\xFF\xFF$}{};
-        $log->tracef(
-            'Size after deflation is %d/%d, ratio of %4.1f%%',
-            length($payload),
-            $original,
-            100.0 * (length($payload) / $original),
+    if($OPCODE_BY_CODE{$type} equ 'close') {
+        my ($code, $reason) = unpack 'na*', $frame->payload;
+        return await $self->close(
+            code   => 1002,
+            reason => 'Invalid UTF-8 reason in close frame',
+        ) unless valid_utf8($reason);
+        await $self->close(
+            code   => ($code || 0),
+            reason => decode_utf8($reason // ''),
         );
     }
-    my $len = length $payload;
-    my $msg = pack('C1', $opcode);
-    if($len < 126) {
-        $msg .= pack('C1', $len);
-    } elsif($len < 0xFFFF) {
-        $msg .= pack('C1n1', 126, $len);
-    } else {
-        $msg .= pack('C1Q>1', 127, $len);
+    return $frame;
+}
+
+async method close (%args) {
+    # Can only close once
+    return if $closed->is_ready;
+
+    # No point trying to write anything if the remote has closed the connection
+    if($stream->is_read_eof) {
+        $closed->done(%args);
+        $stream->close;
+        return;
     }
-    $msg .= $payload;
-    await $stream->write($msg);
-    return;
+
+    my $f = $self->write_frame(
+        type    => 'close',
+        payload => pack(
+            'na*' => ($args{code} // 0), encode_utf8($args{reason} // '')
+        ),
+    );
+    if($server) {
+        $server->on_client_close($self, %args);
+    }
+    $closed->done(%args);
+    await $f;
+    $stream->close;
 }
 
 1;
